#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from pandas import DataFrame

from synnax.channel.payload import (
    ChannelKey,
    ChannelName,
    ChannelParams,
    ChannelPayload,
    normalize_channel_params,
)
from synnax.channel.retrieve import ChannelRetriever, retrieve_required
from synnax.exceptions import ValidationError
from synnax.framer.frame import Frame, CrudeFrame
from synnax.telem.series import CrudeSeries, Series


class ReadFrameAdapter:
    __adapter: dict[ChannelKey, ChannelName] | None
    retriever: ChannelRetriever
    keys: list[ChannelKey]

    def __init__(self, retriever: ChannelRetriever):
        self.retriever = retriever
        self.__adapter = None
        self.keys = list()

    def update(self, channels: ChannelParams):
        normal = normalize_channel_params(channels)
        if normal.variant == "keys":
            self.__adapter = None
            self.keys = normal.channels
            return

        fetched = self.retriever.retrieve(normal.channels)
        self.__adapter = dict[int, str]()
        for name in normal.channels:
            ch = next((c for c in fetched if c.name == name), None)
            if ch is None:
                raise KeyError(f"Channel {name} not found.")
            self.__adapter[ch.key] = ch.name
        self.keys = list(self.__adapter.keys())

    def adapt(self, fr: Frame):
        if self.__adapter is None:
            return fr
        keys = [
            self.__adapter[k] if isinstance(k, ChannelKey) else k for k in fr.channels
        ]
        return Frame(channels=keys, series=fr.series)


class WriteFrameAdapter:
    __adapter: dict[ChannelName, ChannelKey] | None
    retriever: ChannelRetriever
    __keys: list[ChannelKey] | None
    __err_on_extra_chans: bool

    def __init__(self, retriever: ChannelRetriever, err_on_extra_chans: bool = True):
        self.retriever = retriever
        self.__adapter = None
        self.__keys = None
        self.__err_on_extra_chans = err_on_extra_chans

    def update(self, channels: ChannelParams):
        results = retrieve_required(self.retriever, channels)
        self.__adapter = {ch.name: ch.key for ch in results}
        self.__keys = [ch.key for ch in results]

    def adapt_dict_keys(
        self, data: dict[ChannelPayload | ChannelKey | ChannelName, any]
    ) -> dict[ChannelKey, any]:
        out = dict()
        for k in data.keys():
            out[self.__adapt_to_key(k)] = data[k]
        return out

    @property
    def keys(self):
        return self.__keys

    def __adapt_to_key(
        self, ch: ChannelPayload | ChannelKey | ChannelName
    ) -> ChannelKey:
        if isinstance(ch, ChannelKey):
            return ch
        if isinstance(ch, ChannelPayload):
            return ch.key
        # If it's not a payload or key already, it has to be a name,
        # which means we need to resolve the key from a remote source
        # (either cache or server)
        return self.__adapt_ch(ch).key

    def __adapt_ch(
        self, ch: ChannelKey | ChannelName | ChannelPayload
    ) -> ChannelPayload:
        if isinstance(ch, (ChannelKey, ChannelName)):
            return self.retriever.retrieve_one(ch)
        return ch

    def adapt(
        self,
        channels_or_data: ChannelPayload
        | list[ChannelPayload]
        | ChannelParams
        | CrudeFrame,
        series: CrudeSeries | list[CrudeSeries] | None = None,
    ) -> Frame:
        if isinstance(channels_or_data, (ChannelName, ChannelKey, ChannelPayload)):
            if series is None:
                raise ValidationError(
                    f"""
                Received a single channel {'name' if isinstance(channels_or_data, ChannelName) else 'key'}
                but no data.
                """
                )
            if isinstance(series, list) and len(series) > 1:
                first = series[0]
                if not isinstance(first, (float, int)):
                    raise ValidationError(
                        f"""
                    Received a single channel {'name' if isinstance(channels_or_data, ChannelName) else 'key'}
                    but multiple series.
                    """
                    )

            pld = self.__adapt_ch(channels_or_data)
            series = Series(data_type=pld.data_type, data=series)
            return Frame([pld.key], [series])

        if isinstance(channels_or_data, list):
            if series is None:
                raise ValidationError(
                    f"""
                Received {len(channels_or_data)} channels but no series.
                """
                )
            channels = list()
            o_series = list()
            for i, ch in enumerate(channels_or_data):
                pld = self.__adapt_ch(ch)
                if i >= len(series):
                    raise ValidationError(
                        f"""
                    Received {len(channels_or_data)} channels but only {len(series)} series.
                    """
                    )
                s = Series(data_type=pld.data_type, data=series[i])
                channels.append(pld.key)
                o_series.append(s)

            return Frame(channels, o_series)

        is_frame = isinstance(channels_or_data, Frame)
        is_df = isinstance(channels_or_data, DataFrame)
        if is_frame or is_df:
            cols = channels_or_data.channels if is_frame else channels_or_data.columns
            if self.__adapter is None:
                return channels_or_data
            channels = list()
            series = list()
            for col in cols:
                try:
                    channels.append(
                        self.__adapter[col] if isinstance(col, ChannelName) else col
                    )
                    series.append(Series(channels_or_data[col]))
                except KeyError as e:
                    if self.__err_on_extra_chans:
                        raise ValidationError(
                            f"Channel {e} was not provided in the list of "
                            f"channels when the writer was opened."
                        )
            return Frame(channels=channels, series=series)

        if isinstance(channels_or_data, dict):
            channels = list()
            series = list()
            for k, v in channels_or_data.items():
                pld = self.__adapt_ch(k)
                s = Series(data_type=pld.data_type, data=v)
                channels.append(pld.key)
                series.append(s)

            return Frame(channels, series)

        raise TypeError(
            f"""Cannot construct frame from {channels_or_data} and {series}"""
        )



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from uuid import UUID
from typing import overload

from alamos import Instrumentation, NOOP
from freighter import Payload, UnaryClient, send_required, Empty
from synnax.access.payload import Policy
from synnax.ontology.payload import ID
from synnax.util.normalize import normalize


class _CreateRequest(Payload):
    policies: list[Policy]


_CreateResponse = _CreateRequest


class _RetrieveRequest(Payload):
    keys: list[UUID] | None
    subjects: list[ID] | None


class _RetrieveResponse(Payload):
    policies: list[Policy] | None


class _DeleteRequest(Payload):
    keys: list[UUID]


ONTOLOGY_TYPE = ID(type="policy")

_CREATE_ENDPOINT = "/access/policy/create"
_RETRIEVE_ENDPOINT = "/access/policy/retrieve"
_DELETE_ENDPOINT = "/access/policy/delete"


class PolicyClient:
    _client: UnaryClient
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ):
        self._client = client
        self.instrumentation = instrumentation

    @overload
    def create(
        self,
        *,
        subjects: list[ID] = None,
        objects: list[ID] = None,
        actions: list[str] = None,
    ) -> Policy:
        ...

    @overload
    def create(
        self,
        policies: Policy,
    ) -> Policy:
        ...

    @overload
    def create(
        self,
        policies: list[Policy],
    ) -> list[Policy]:
        ...

    def create(
        self,
        policies: Policy | list[Policy] | None = None,
        *,
        subjects: list[ID] = None,
        objects: list[ID] = None,
        actions: list[str] = None,
    ) -> Policy | list[Policy]:
        is_single = not isinstance(policies, list)
        if policies is None:
            policies = Policy(
                subjects=subjects,
                objects=objects,
                actions=actions,
            )
        req = _CreateRequest(policies=normalize(policies))
        res = send_required(self._client, _CREATE_ENDPOINT, req, _CreateResponse)
        return res.policies[0] if is_single else res.policies

    def retrieve(
        self, keys: list[UUID] | None = None, subjects: list[ID] | None = None
    ) -> list[Policy]:
        res = send_required(
            self._client,
            _RETRIEVE_ENDPOINT,
            _RetrieveRequest(keys=keys, subjects=subjects),
            _RetrieveResponse,
        )
        return [] if res is None else res.policies

    def delete(self, keys: UUID | list[UUID]) -> None:
        req = _DeleteRequest(keys=normalize(keys))
        send_required(self._client, _DELETE_ENDPOINT, req, Empty)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import uuid

from typing import overload
from freighter import Payload, UnaryClient, send_required

from synnax.util.normalize import normalize


class KVPair(Payload):
    range: uuid.UUID
    key: str
    value: str


class _GetRequest(Payload):
    range: uuid.UUID
    keys: list[str]


class _GetResponse(Payload):
    pairs: list[KVPair]


class _SetRequest(Payload):
    range: uuid.UUID
    pairs: list[KVPair]


class _DeleteRequest(Payload):
    range: uuid.UUID
    keys: list[str]


class _EmptyResponse(Payload):
    ...


_SET_ENDPOINT = "/range/kv/set"
_GET_ENDPOINT = "/range/kv/get"
_DELETE_ENDPOINT = "/range/kv/delete"


class KV:
    _client: UnaryClient
    _rng_key: uuid.UUID

    def __init__(self, rng: uuid.UUID, client: UnaryClient) -> None:
        self._client = client
        self._rng_key = rng

    @overload
    def get(self, keys: str) -> str:
        ...

    def get(self, keys: str | list[str]) -> dict[str, str] | str:
        req = _GetRequest(range=self._rng_key, keys=normalize(keys))
        res = send_required(self._client, _GET_ENDPOINT, req, _GetResponse)
        if isinstance(keys, str):
            return res.pairs[0].value
        return {pair.key: pair.value for pair in res.pairs}

    @overload
    def set(self, key: str, value: str):
        ...

    @overload
    def set(self, key: dict[str, str]):
        ...

    def set(self, key: str | dict[str, str], value: str | None = None) -> None:
        pairs = list()
        if isinstance(key, str):
            pairs.append(KVPair(range=self._rng_key, key=key, value=value))
        else:
            for k, v in key.items():
                pairs.append(KVPair(range=self._rng_key, key=k, value=v))
        req = _SetRequest(range=self._rng_key, pairs=pairs)
        send_required(self._client, _SET_ENDPOINT, req, _EmptyResponse)

    def delete(self, keys: str | list[str]) -> None:
        req = _DeleteRequest(range=self._rng_key, keys=normalize(keys))
        send_required(self._client, _DELETE_ENDPOINT, req, _EmptyResponse)

    def __getitem__(self, key: str) -> str:
        return self.get(key)

    def __setitem__(self, key: str, value: str) -> None:
        self.set(key, value)

    def __delitem__(self, key: str) -> None:
        self.delete(key)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from typing import overload
from numpy import ndarray
from pydantic import PrivateAttr

from synnax.channel.payload import (
    ChannelKey,
    ChannelKeys,
    ChannelName,
    ChannelNames,
    ChannelParams,
    ChannelPayload,
    normalize_channel_params,
)
from synnax.channel.retrieve import ChannelRetriever
from synnax.channel.writer import ChannelWriter
from synnax.exceptions import NotFoundError, MultipleFoundError, ValidationError
from synnax.framer.client import Client as FrameClient
from synnax.ontology.payload import ID
from synnax.telem import (
    CrudeDataType,
    CrudeRate,
    CrudeTimeStamp,
    DataType,
    Rate,
    Series,
    TimeRange,
)
from synnax.util.normalize import normalize

CHANNEL_ONTOLOGY_TYPE = ID(type="channel")


class Channel(ChannelPayload):
    """A channel is a logical collection of samples emitted by or representing the
    values of a single source. See
    https://docs.synnaxlabs.com/reference/concepts/channels for an introduction to
    channels and how they work.
    """

    ___frame_client: FrameClient | None = PrivateAttr(None)
    __client: ChannelClient | None = PrivateAttr(None)

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self,
        *,
        name: str,
        data_type: CrudeDataType,
        rate: CrudeRate = 0,
        is_index: bool = False,
        index: ChannelKey = 0,
        leaseholder: int = 0,
        key: ChannelKey = 0,
        virtual: bool = False,
        internal: bool = False,
        _frame_client: FrameClient | None = None,
        _client: ChannelClient | None = None,
    ) -> None:
        """Initializes a new Channel using the given parameters. It's important to note
        that this does not create the Channel in the cluster. To create the channel,
        call client.channels.create(channel).

        :param data_type: The data type of the samples in the channel e.g. np.int64
        :param rate: Rate sets the rate at which the channels values are written. If
        this parameter is non-zero, is_index must be false and index must be an empty
        string or unspecified.
        :param name: A human-readable name for the channel.
        :param key: Is auto-assigned by the cluster, and should not be set by the
        caller.
        :param is_index: Boolean indicating whether the channel is an index. Index
        channels should have ax data type of synnax.TIMESTAMP.
        :param index: The key of the channel that indexes this channel.
        :param leaseholder: The node that holds the lease for this channel. If you
        don't know what this is, leave it at the default value of 0.
        :param _frame_client: The backing client for reading and writing data to and
        from the channel. This is provided by the Synnax py during calls to
        .channels.create() and .channels.retrieve() and should not be set by the caller.
        """
        super().__init__(
            data_type=DataType(data_type),
            rate=Rate(rate),
            name=name,
            leaseholder=leaseholder,
            key=key,
            is_index=is_index,
            index=index,
            internal=internal,
            virtual=virtual,
        )
        self.___frame_client = _frame_client
        self.__client = _client

    @overload
    def read(
        self,
        start_or_range: TimeRange,
    ) -> Series:
        ...

    @overload
    def read(
        self,
        start_or_range: CrudeTimeStamp,
        end: CrudeTimeStamp,
    ) -> Series:
        ...

    def read(
        self,
        start_or_range: CrudeTimeStamp | TimeRange,
        end: CrudeTimeStamp | None = None,
    ) -> Series:
        """Reads telemetry from the channel between the two timestamps.

        :param start_or_range: The starting timestamp of the range to read from.
        :param end: The ending timestamp of the range to read from.
        :returns: A tuple containing a numpy array of the telemetry and a TimeRange
        representing the range of telemetry. The start of the time range represents
        the timestamp of the first sample in the array.
        :raises ContiguityError: If the telemetry between start and end is non-contiguous.
        """
        tr = TimeRange(start_or_range, end)
        return self.__frame_client.read(tr, self.key)

    def write(self, start: CrudeTimeStamp, data: ndarray | Series) -> None:
        """Writes telemetry to the channel starting at the given timestamp.

        :param start: The starting timestamp of the first sample in data.
        :param data: The telemetry to write to the channel.
        :returns: None.
        """
        self.__frame_client.write(start, self.key, data)

    def rename(self, name: ChannelName) -> None:
        """Renames the channel.

        :param name: The new name for the channel.
        :returns: None.
        """
        self.__client.rename(self.key, name)

    @property
    def __frame_client(self) -> FrameClient:
        if self.___frame_client is None:
            raise ValidationError(
                "Cannot read from or write to channel that has not been created."
            )
        return self.___frame_client

    def __hash__(self) -> int:
        return hash(self.key)

    def __eq__(self, other) -> bool:
        return self.key == other.key

    def to_payload(self) -> ChannelPayload:
        return ChannelPayload(
            data_type=self.data_type,
            rate=self.rate,
            name=self.name,
            leaseholder=self.leaseholder,
            key=self.key,
            index=self.index,
            is_index=self.is_index,
        )


class ChannelClient:
    """The core py class for executing channel operations against a Synnax cluster."""

    _frame_client: FrameClient
    _retriever: ChannelRetriever
    _creator: ChannelWriter

    def __init__(
        self,
        frame_client: FrameClient,
        retriever: ChannelRetriever,
        creator: ChannelWriter,
    ):
        self._frame_client = frame_client
        self._retriever = retriever
        self._creator = creator

    def delete(self, channels: ChannelParams) -> None:
        """Deletes on or more channels from the cluster"""
        self._creator.delete(channels)

    @overload
    def create(
        self,
        *,
        data_type: CrudeDataType = DataType.UNKNOWN,
        name: ChannelName = "",
        rate: CrudeRate = Rate(0),
        index: ChannelKey = 0,
        is_index: bool = False,
        leaseholder: int = 0,
        retrieve_if_name_exists: bool = False,
    ) -> Channel:
        ...

    @overload
    def create(
        self, channels: Channel, *, retrieve_if_name_exists: bool = False
    ) -> Channel:
        ...

    @overload
    def create(
        self, channels: list[Channel], *, retrieve_if_name_exists: bool = False
    ) -> list[Channel]:
        ...

    def create(
        self,
        channels: Channel | list[Channel] | None = None,
        *,
        data_type: CrudeDataType = DataType.UNKNOWN,
        name: ChannelName = "",
        rate: CrudeRate = Rate(0),
        is_index: bool = False,
        index: ChannelKey = 0,
        leaseholder: int = 0,
        virtual: bool = False,
        retrieve_if_name_exists: bool = False,
    ) -> Channel | list[Channel]:
        """Creates a new channel or set of channels in the cluster. Possible arguments
        are as follows:

        Overload 1:
        :param data_type: The data type of the samples in the channel e.g np.int64
        :param rate: Rate sets the rate at which the channels values are written. If this
        parameter is non-zero, is_index must be false and index must be an empty string or
        unspecified.
        :param name: A human-readable name for the channel.
        :param is_index: Boolean indicating whether the channel is an index. Index
        channels should have ax data type of synnax.TIMESTAMP.
        :param index: The key or channel that indexes this channel.
        :param leaseholder: The node that holds the lease for this channel. If you don't know
        what this is, leave it at the default value of 0.
        :returns: The created channel.

        Overload 2:

        :param channels: A single channel to create.
        :returns: The created channel.

        Overload 3:

        :param channels: A list of channels to create.
        :returns: The created channels.
        """

        if channels is None:
            if is_index and data_type == DataType.UNKNOWN:
                data_type = DataType.TIMESTAMP
            _channels = [
                ChannelPayload(
                    name=name,
                    leaseholder=leaseholder,
                    rate=Rate(rate),
                    data_type=DataType(data_type),
                    index=index,
                    is_index=is_index,
                    virtual=virtual,
                )
            ]
        elif isinstance(channels, Channel):
            _channels = [channels.to_payload()]
        else:
            _channels = [c.to_payload() for c in channels]

        created = list()
        if retrieve_if_name_exists:
            created = self.__sugar(
                self._retriever.retrieve([ch.name for ch in _channels])
            )
            _channels = [
                c for c in _channels if c.name not in [ch.name for ch in created]
            ]

        created.extend(self.__sugar(self._creator.create(_channels)))
        return created if isinstance(channels, list) else created[0]

    @overload
    def retrieve(self, channel: ChannelKey | ChannelName) -> Channel:
        ...

    @overload
    def retrieve(
        self,
        channel: ChannelKeys | ChannelNames,
    ) -> list[Channel]:
        ...

    def retrieve(self, channel: ChannelParams) -> Channel | list[Channel]:
        """Retrieves a channel or set of channels from the cluster.

        Overload 1:

        :param key: The key of the channel to retrieve. If this is specified, the name
        parameter is ignored.
        :param name: The name of the channel to retrieve. If key is specified, this is
        ignored.
        Only one of key or name must be specified.
        :returns: The associated channel.
        :raises QueryError: If the channel is not found.

        Overload 2 + 3:
        :param keys: The keys of the channels to retrieve. If this is specified, the names
        parameter is ignored.
        :param names: The names of the channels to retrieve. If keys are specified, this is
        ignored.
        Only one of keys or names may be specified.
        :param leaseholder: The node that holds the lease for the channels to retrieve. If you
        don't know what this is, don't specify it.
        :param include_not_found: Boolean indicating whether or not to include the keys or
        names of the channels that were not found in the result.
        :returns: The retrieved channels if include_not_found is False, otherwise a tuple
        containing the retrieved channels and the keys or names of the channels that were
        not found.
        """
        normal = normalize_channel_params(channel)
        res = self._retriever.retrieve(channel)
        sug = self.__sugar(res)
        if not normal.single:
            return sug

        if len(res) == 1:
            return sug[0]

        if len(res) > 1:
            raise _multiple_results_error(channel, res)

        raise NotFoundError(f"Channel matching '{channel}' not found.")

    @overload
    def rename(self, keys: ChannelKey, names: ChannelName) -> None:
        """Renames a channel in the cluster.

        :param keys: The key of the channel to rename.
        :param names: The new name for the channel.
        :returns: None.
        """
        ...

    @overload
    def rename(self, keys: ChannelKeys, names: ChannelNames) -> None:
        """Renames one or more channels in the cluster.

        :param keys: The keys of the channels to rename.
        :param names: The new names for the channels.
        :returns: None.
        """

    def rename(self, keys: ChannelKeys, names: ChannelNames) -> None:
        """Renames one or more channels in the cluster.

        :param keys: The keys of the channels to rename.
        :param names: The new names for the channels.
        :returns: None.
        """
        self._creator.rename(normalize(keys), normalize(names))

    def __sugar(self, channels: list[ChannelPayload]) -> list[Channel]:
        return [Channel(**c.dict(), _frame_client=self._frame_client) for c in channels]


def _multiple_results_error(
    channel: ChannelParams,
    results: list[ChannelPayload],
) -> MultipleFoundError:
    msg = f"""

{len(results)} channels matching '{channel}' found. If you'd like to retrieve all
of them, pass in '{channel}' as an array i.e. ['{channel}'] instead of {channel}.

The channels found were:
    """

    # append a max of five results to the error message. If we have more than five,
    # we'll just say "and x more" at the end.
    for i in range(min(5, len(results))):
        msg += f"{str(results[i])}, "

    if len(results) > 5:
        msg += f"and {len(results) - 5} more."

    return MultipleFoundError(msg)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


from __future__ import annotations

import functools
from typing import Callable, overload

from freighter import UnaryClient

from synnax.channel.retrieve import ChannelRetriever
from synnax.exceptions import QueryError
from synnax.framer.client import Client
from synnax.framer.frame import CrudeFrame
from synnax.ranger.alias import Aliaser
from synnax.ranger.kv import KV
from synnax.ranger.payload import (
    RangeKey,
    RangeKeys,
    RangeName,
    RangeNames,
    RangePayload,
)
from synnax.ranger.retrieve import RangeRetriever
from synnax.ranger.writer import RangeWriter
from synnax.signals.signals import Registry
from synnax.state import LatestState
from synnax.telem import (
    TimeRange,
    Series,
    SampleValue,
    DataType,
    Rate,
    CrudeSeries,
)
from synnax.ontology.payload import ID
from synnax.channel.payload import (
    ChannelKey,
    ChannelName,
    ChannelPayload,
    ChannelKeys,
    ChannelNames,
    ChannelParams,
)
from synnax.util.interop import overload_comparison_operators
from synnax.hardware.task import Task, Client as TaskClient
from synnax.ontology import Client as OntologyClient
from synnax.hardware.ni import AnalogReadTask
from synnax.util.normalize import check_for_none, normalize

from uuid import UUID
from typing import overload

import numpy as np
from pydantic import PrivateAttr

RANGE_SET_CHANNEL = "sy_range_set"


class _InternalScopedChannel(ChannelPayload):
    __range: Range | None = PrivateAttr(None)
    """The range that this channel belongs to."""
    __frame_client: Client | None = PrivateAttr(None)
    """The frame client for executing read operations."""
    __aliaser: Aliaser | None = PrivateAttr(None)
    """An aliaser for setting the channel's alias."""
    __cache: Series | None = PrivateAttr(None)
    """An internal cache to prevent repeated reads from the same channel."""
    __tasks: TaskClient | None = PrivateAttr(None)
    __ontology: OntologyClient | None = PrivateAttr(None)

    def __new__(cls, *args, **kwargs):
        cls = overload_comparison_operators(cls, "__array__")
        return super().__new__(cls)

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self,
        rng: Range,
        frame_client: Client,
        tasks: TaskClient,
        ontology: OntologyClient,
        payload: ChannelPayload,
        aliaser: Aliaser | None = None,
    ):
        super().__init__(**payload.dict())
        self.__range = rng
        self.__frame_client = frame_client
        self.__aliaser = aliaser
        self.__tasks = tasks
        self.__ontology = ontology

    @property
    def time_range(self) -> TimeRange:
        return self.__range.time_range

    @property
    def calibrations(self):
        snapshots = self.__range.snapshots()
        ni_tasks = [AnalogReadTask(t) for t in snapshots]
        for t in ni_tasks:
            for chan in t.config.channels:
                if chan.channel == self.key:
                    return chan

        return None

    def __array__(self, *args, **kwargs) -> np.ndarray:
        """Converts the channel to a numpy array. This method is necessary
        for numpy interop."""
        return self.read().__array__(*args, **kwargs)

    def __getitem__(self, index: int) -> SampleValue:
        return self.read().__getitem__(index)

    def to_numpy(self) -> np.ndarray:
        """Converts the channel to a numpy array. This method is necessary
        for matplotlib interop.
        """
        return self.read().to_numpy()

    def read(self) -> Series:
        if self.__cache is None:
            self.__cache = self.__frame_client.read(self.time_range, self.key)
        return self.__cache

    def set_alias(self, alias: str):
        self.__range.set_alias(self.key, alias)

    def __str__(self) -> str:
        return f"{super().__str__()} between {self.time_range.start} and {self.time_range.end}"

    def __len__(self):
        return len(self.read())


class ScopedChannel:
    """A channel that is scoped to a particular range. This class is returned when
    accessing the channel as a key or property on a range. This channel has direct
    interoperability with numpy arrays, meaning that it can be passed as an argument
    to any function/method that accepts a numpy array.

    It's very important to note that if the property accessor matches multiple channels,
    this class will contain all of them, and, as a result, single channel operations
    will fail. However, you can use the __iter__ method to iterate over the channels
    in the returned values. This is particularly relevant when using regex to match
    multiple channels.
    """

    __internal: list[_InternalScopedChannel]
    __query: str

    def __new__(cls, *args, **kwargs):
        cls = overload_comparison_operators(cls, "__array__")
        return super().__new__(cls)

    def __init__(
        self,
        query: str,
        internal: list[_InternalScopedChannel],
    ):
        self.__internal = internal
        self.__query = query

    def __guard(self):
        if len(self.__internal) > 1:
            raise QueryError(
                f"""Multiple channels found for query '{self.__query}':
            {[str(ch) for ch in self.__internal]}
            """
            )

    def __array__(self, *args, **kwargs):
        """Converts the scoped channel to a numpy array. This method is necessary
        for numpy interop."""
        self.__guard()
        return self.__internal[0].__array__(*args, **kwargs)

    def __getitem__(self, index: int) -> SampleValue:
        self.__guard()
        return self.__internal[0].__getitem__(index)

    def to_numpy(self) -> np.ndarray:
        """Converts the scoped channel to a numpy array. This method is necessary
        for matplotlib interop."""
        return self.__array__()

    @property
    def key(self) -> ChannelKey:
        self.__guard()
        return self.__internal[0].key

    @property
    def name(self) -> str:
        self.__guard()
        return self.__internal[0].name

    @property
    def data_type(self) -> DataType:
        self.__guard()
        return self.__internal[0].data_type

    @property
    def is_index(self) -> bool:
        self.__guard()
        return self.__internal[0].is_index

    @property
    def index(self) -> ChannelKey:
        self.__guard()
        return self.__internal[0].index

    @property
    def leaseholder(self) -> int:
        self.__guard()
        return self.__internal[0].leaseholder

    @property
    def rate(self) -> Rate:
        self.__guard()
        return self.__internal[0].rate

    @property
    def calibrations(self):
        self.__guard()
        return self.__internal[0].calibrations

    def set_alias(self, alias: str):
        self.__guard()
        self.__internal[0].set_alias(alias)

    def __iter__(self):
        return iter(self.__internal)

    def __len__(self):
        return sum(len(ch) for ch in self.__internal)


_RANGE_NOT_CREATED = QueryError(
    """Cannot read from a range that has not been created.
Please call client.ranges.create(range) before attempting to read from a range."""
)


class Range(RangePayload):
    """A range is a user-defined region of a cluster's data. It's identified by a name,
    time range, and uniquely generated key. See
    https://docs.synnaxlabs.com/reference/concepts/ranges for an introduction to ranges
    and how they work.
    """

    __frame_client: Client | None = PrivateAttr(None)
    """The frame client for executing read and write operations."""
    _channels: ChannelRetriever | None = PrivateAttr(None)
    """For retrieving channels from the cluster."""
    _kv: KV | None = PrivateAttr(None)
    """Key-value store for storing metadata about the range."""
    __aliaser: Aliaser | None = PrivateAttr(None)
    """For setting and resolving aliases."""
    _cache: dict[ChannelKey, _InternalScopedChannel] = PrivateAttr(dict())
    """A writer for creating child ranges"""
    _client: RangeClient | None = PrivateAttr(None)
    _tasks: TaskClient | None = PrivateAttr(None)
    _ontology: OntologyClient | None = PrivateAttr(None)

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self,
        name: str,
        time_range: TimeRange,
        key: UUID = UUID(int=0),
        color: str = "",
        *,
        _frame_client: Client | None = None,
        _channel_retriever: ChannelRetriever | None = None,
        _kv: KV | None = None,
        _aliaser: Aliaser | None = None,
        _client: RangeClient | None = None,
        _tasks: TaskClient | None = None,
        _ontology: OntologyClient | None = None,
    ):
        """Initializes a new Range using the given parameters. It's important to note
        that this does not create the Range in the cluster. To create the range, call
        client.ranges.create(range).

        :param name: A human-readable name for the range. This should represent the data
            that the range contains i.e. "Hotfire 1", "Print 22", or "Tank Burst Test.".
        :param time_range: The time region spanned by the range. Note that this range
            is end inclusive and end exclusive i.e. the start represents the timestamp
            of just before or at the first data point in the range, and the end
            represents the timestamp of the just after the last data point in the range.
        :param key: A UUID that uniquely identifies the range. This is typically not
            set by the user, and is generated by the cluster upon creating the range.
        :param _frame_client: The backing client for reading and writing data to
            and from the cluster. This is provided by Synnax during calls to
            .ranges.create() and .ranges.retrieve(), and should not be set by the user.
        :param _channel_retriever: The backing client for retrieving channels to
            and from the cluster. This is provided by Synnax during calls to
            .ranges.create() and .ranges.retrieve(), and should not be set by the user.
        """
        super().__init__(name=name, time_range=time_range, key=key, color=color)
        self.__frame_client = _frame_client
        self._channels = _channel_retriever
        self._kv = _kv
        self.__aliaser = _aliaser
        self._client = _client
        self._tasks = _tasks
        self._ontology = _ontology

    def __getattr__(self, query: str) -> ScopedChannel:
        channels = self._channel_retriever.retrieve(query)
        aliases = self._aliaser.resolve([query])
        channels.extend(self._channel_retriever.retrieve(list(aliases.values())))
        if len(channels) == 0:
            raise QueryError(f"Channel matching {query} not found")

        return ScopedChannel(query, self.__splice_cached(channels))

    def __getitem__(self, name: str | ChannelKey) -> ScopedChannel:
        return self.__getattr__(name)

    def __splice_cached(
        self, channels: list[ChannelPayload]
    ) -> list[_InternalScopedChannel]:
        results = list()
        for pld in channels:
            cached = self._cache.get(pld.key, None)
            if cached is None:
                cached = _InternalScopedChannel(
                    rng=self,
                    frame_client=self._frame_client,
                    payload=pld,
                    tasks=self._tasks,
                    ontology=self._ontology,
                )
                self._cache[pld.key] = cached
            results.append(cached)
        return results

    @property
    def ontology_id(self) -> ID:
        return ID(type="range", key=str(self.key))

    @property
    def meta_data(self):
        if self._kv is None:
            raise _RANGE_NOT_CREATED
        return self._kv

    @property
    def _aliaser(self):
        if self.__aliaser is None:
            raise _RANGE_NOT_CREATED
        return self.__aliaser

    @property
    def _frame_client(self) -> Client:
        if self.__frame_client is None:
            raise _RANGE_NOT_CREATED
        return self.__frame_client

    @property
    def _channel_retriever(self) -> ChannelRetriever:
        if self._channels is None:
            raise _RANGE_NOT_CREATED
        return self._channels

    @overload
    def set_alias(self, channel: ChannelKey | ChannelName, alias: str):
        ...

    @overload
    def set_alias(self, channel: dict[ChannelKey | ChannelName, str]):
        ...

    def set_alias(
        self,
        channel: ChannelKey | ChannelName | dict[ChannelKey | ChannelName, str],
        alias: str = None,
    ):
        if not isinstance(channel, dict):
            if alias is None:
                raise ValueError("Alias must be provided if channel is not a dict")
            channel = {channel: alias}
        corrected = {}
        for ch, alias in channel.items():
            if isinstance(ch, ChannelName):
                res = self._channel_retriever.retrieve(ch)
                if len(res) == 0:
                    raise QueryError(f"Channel {ch} not found")
                corrected[res[0].key] = alias
            else:
                corrected[ch] = alias
        self._aliaser.set(corrected)

    def to_payload(self) -> RangePayload:
        return RangePayload(name=self.name, time_range=self.time_range, key=self.key)

    @overload
    def write(self, to: ChannelKey | ChannelName | ChannelPayload, data: CrudeSeries):
        ...

    @overload
    def write(
        self,
        to: ChannelKeys | ChannelNames | list[ChannelPayload],
        series: list[CrudeSeries],
    ):
        ...

    @overload
    def write(self, frame: CrudeFrame):
        ...

    def write(
        self,
        to: ChannelParams | ChannelPayload | list[ChannelPayload] | CrudeFrame,
        series: CrudeSeries | list[CrudeSeries] | None = None,
    ):
        self.__frame_client.write(self.time_range.start, to, series)

    def create_sub_range(
        self,
        *,
        name: str,
        time_range: TimeRange,
        color: str = "",
        key: RangeKey = UUID(int=0),
    ) -> Range:
        return self._client.create(
            name=name,
            time_range=time_range,
            color=color,
            key=key,
            parent=ID(type="range", key=str(self.key)),
        )

    def snapshots(self) -> list[Task]:
        res = self._ontology.retrieve_children(self.ontology_id)
        tasks = [t for t in res if t.id.type == "task"]
        return self._tasks.retrieve(keys=[t.id.key for t in tasks])


class RangeClient:
    _frame_client: Client
    _channels: ChannelRetriever
    _retriever: RangeRetriever
    _writer: RangeWriter
    _unary_client: UnaryClient
    _signals: Registry
    _tasks: TaskClient
    _ontology: OntologyClient

    def __init__(
        self,
        unary_client: UnaryClient,
        frame_client: Client,
        writer: RangeWriter,
        retriever: RangeRetriever,
        channel_retriever: ChannelRetriever,
        signals: Registry,
        tasks: TaskClient,
        ontology: OntologyClient,
    ) -> None:
        self._unary_client = unary_client
        self._frame_client = frame_client
        self._writer = writer
        self._retriever = retriever
        self._channels = channel_retriever
        self._signals = signals
        self._tasks = tasks
        self._ontology = ontology

    @overload
    def create(
        self,
        *,
        name: str,
        time_range: TimeRange,
        color: str = "",
        retrieve_if_name_exists: bool = False,
        parent: ID | None = None,
        key: RangeKey = UUID(int=0),
    ) -> Range:
        """Creates a named range spanning a region of time. This range is persisted
        to the cluster and is visible to all clients.

        :param name: The name of the range
        :param time_range: The time range of the range. This time range must be valid
        i.e. its start time must be less than or equal to its end time.
        :param retrieve_if_name_exists: If true, this method will retrieve the range
        if it already exists. This method will throw a ValidationError
        if the range already exists and has a DIFFERENT time range.
        :param parent: An optional parent ontology item to set as the parent of the
        range.
        """
        ...

    @overload
    def create(
        self,
        ranges: Range,
        retrieve_if_name_exists: bool = False,
        parent: ID | None = None,
    ) -> Range:
        """Creates the given range. This range is persisted to the cluster and is
        visible to all clients.

        :param ranges: The range to create
        :param retrieve_if_name_exists: If true, this method will retrieve the range
        if it already exists. This method will throw a ValidationError if the range
        already exists and has a DIFFERENT time range.
        :param parent: An optional parent ontology item to set as the parent of the
        range.
        """
        ...

    @overload
    def create(
        self,
        ranges: list[Range],
        retrieve_if_name_exists: bool = False,
        parent: ID | None = None,
    ) -> list[Range]:
        """Creates the given ranges. These ranges are persisted to the cluster and are
        visible to all clients.

        :param ranges: The ranges to create
        :param retrieve_if_name_exists: If true, this method will retrieve the range
        if it already exists. This method will throw a ValidationError if the range
        already exists and has a DIFFERENT time range.
        :param parent: An optional parent ontology item to set as the parent of the
        """
        ...

    def create(
        self,
        ranges: Range | list[Range] = None,
        *,
        key: RangeKey = UUID(int=0),
        name: str = "",
        time_range: TimeRange | None = None,
        color: str = "",
        retrieve_if_name_exists: bool = False,
        parent: ID | None = None,
    ) -> Range | list[Range]:
        is_single = True
        if ranges is None:
            to_create = [
                RangePayload(key=key, name=name, time_range=time_range, color=color)
            ]
        elif isinstance(ranges, Range):
            to_create = [ranges.to_payload()]
        else:
            is_single = False
            to_create = [r.to_payload() for r in ranges]

        res: list[Range] = list()
        if retrieve_if_name_exists:
            res = self.retrieve(names=[r.name for r in to_create])
            if is_single and len(res) > 1:
                filtered = [r for r in res if r.time_range == time_range]
                if len(filtered) == 0:
                    raise QueryError(
                        f"""
                        retrieve_if_name_exists was set to true, but {len(res)} ranges
                        were found matching {name} but none had the same time range as
                        passed to create. Synnax can't figure out which one you want!
                        """
                    )
                if len(filtered) > 1:
                    raise QueryError(
                        f"""
                    retrieve_if_name_exists was set to true, but {len(res)} ranges were
                    found that matched {name} and had time range {time_range}. Synnax
                    can't figure out which one you want!
                    """
                    )
                res = [filtered[0]]
            existing_names = {r.name for r in res}
            to_create = [r for r in to_create if r.name not in existing_names]
        if len(to_create) > 0:
            res.extend(self.__sugar(self._writer.create(to_create, parent=parent)))
        return res if not is_single else res[0]

    @overload
    def retrieve(self, *, key: RangeKey) -> Range:
        ...

    @overload
    def retrieve(self, *, name: RangeName) -> Range:
        ...

    @overload
    def retrieve(self, *, names: RangeNames) -> list[Range]:
        ...

    @overload
    def retrieve(self, *, keys: RangeKeys) -> list[Range]:
        ...

    def retrieve(
        self,
        *,
        key: RangeKey | None = None,
        name: RangeName | None = None,
        names: RangeNames | None = None,
        keys: RangeKeys | None = None,
    ) -> Range | list[Range]:
        is_single = check_for_none(keys, names)
        _ranges = self._retriever.retrieve(key=key, name=name, names=names, keys=keys)
        sug = self.__sugar(_ranges)
        if not is_single:
            return sug
        if len(sug) == 0:
            raise QueryError(f"Range {names} not found")
        elif len(sug) > 1:
            raise QueryError(f"Multiple ranges matching {names} found")
        return sug[0]

    def delete(
        self,
        key: RangeKey | RangeKeys,
    ) -> None:
        self._writer.delete(normalize(key))

    def search(
        self,
        term: str,
    ) -> list[Range]:
        _ranges = self._retriever.search(term)
        return self.__sugar(_ranges)

    def __sugar(self, ranges: list[RangePayload]):
        return [
            Range(
                **r.dict(),
                _frame_client=self._frame_client,
                _channel_retriever=self._channels,
                _kv=KV(r.key, self._unary_client),
                _aliaser=Aliaser(r.key, self._unary_client),
                _client=self,
                _ontology=self._ontology,
                _tasks=self._tasks,
            )
            for r in ranges
        ]

    def on_create(self, f: Callable[[Range], None]):
        @functools.wraps(f)
        def wrapper(state: LatestState):
            d = state[RANGE_SET_CHANNEL]
            f(
                Range(
                    name=d["name"],
                    key=d["key"],
                    time_range=TimeRange(
                        start=d["time_range"]["start"],
                        end=d["time_range"]["end"],
                    ),
                    _frame_client=self._frame_client,
                    _channel_retriever=self._channels,
                    _kv=KV(d["key"], self._unary_client),
                    _aliaser=Aliaser(d["key"], self._unary_client),
                )
            )

        return self._signals.on([RANGE_SET_CHANNEL], lambda s: True)(wrapper)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from uuid import UUID
from freighter import UnaryClient, send_required, Payload, Empty
from typing import overload
from synnax.user.payload import NewUser, User
from synnax.util.normalize import normalize


class _CreateRequest(Payload):
    users: list[NewUser]


class _CreateResponse(Payload):
    users: list[User]


class _RetrieveRequest(Payload):
    keys: list[UUID] | None = None
    usernames: list[str] | None = None


class _RetrieveResponse(Payload):
    users: list[User] | None = None


class _DeleteRequest(Payload):
    keys: list[UUID]


class _ChangeUsernameRequest(Payload):
    key: UUID
    username: str


class _ChangeNameRequest(Payload):
    key: UUID
    first_name: str
    last_name: str


_CREATE_ENDPOINT = "/user/create"
_CHANGE_USERNAME_ENDPOINT = "/user/change_username"
_CHANGE_NAME_ENDPOINT = "/user/change_name"
_DELETE_ENDPOINT = "/user/delete"
_RETRIEVE_ENDPOINT = "/user/retrieve"


class Client:
    client: UnaryClient

    def __init__(self, transport: UnaryClient) -> None:
        self.client = transport

    @overload
    def create(
        self,
        *,
        username: str,
        password: str,
        first_name: str = "",
        last_name: str = "",
        key: UUID | None = None,
    ) -> User:
        ...

    @overload
    def create(self, *, user: NewUser) -> User:
        ...

    @overload
    def create(self, *, users: list[NewUser]) -> list[User]:
        ...

    def create(
        self,
        *,
        username: str | None = None,
        password: str | None = None,
        first_name: str | None = None,
        last_name: str | None = None,
        key: UUID | None = None,
        user: NewUser | None = None,
        users: list[NewUser] | None = None,
    ) -> User | list[User]:
        if username is not None:
            if first_name is None:
                first_name = ""
            if last_name is None:
                last_name = ""
            user = NewUser(
                username=username,
                password=password,
                first_name=first_name,
                last_name=last_name,
                key=key,
            )
        single = user is not None
        if single:
            users = [user]
        res = send_required(
            self.client,
            _CREATE_ENDPOINT,
            _CreateRequest(users=users),
            _CreateResponse,
        ).users
        if single:
            return res[0]
        return res

    def change_username(self, key: UUID, username: str) -> None:
        send_required(
            self.client,
            _CHANGE_USERNAME_ENDPOINT,
            _ChangeUsernameRequest(key=key, username=username),
            Empty,
        )

    def change_name(
        self, key: UUID, *, first_name: str = "", last_name: str = ""
    ) -> None:
        send_required(
            self.client,
            _CHANGE_NAME_ENDPOINT,
            _ChangeNameRequest(key=key, first_name=first_name, last_name=last_name),
            Empty,
        )

    @overload
    def retrieve(self, *, key: UUID) -> User:
        ...

    @overload
    def retrieve(self, *, keys: list[UUID]) -> list[User]:
        ...

    @overload
    def retrieve(self, *, username: str) -> User:
        ...

    @overload
    def retrieve(self, *, usernames: list[str]) -> list[User]:
        ...

    def retrieve(
        self,
        *,
        key: UUID | None = None,
        keys: list[UUID] | None = None,
        username: str | None = None,
        usernames: list[str] | None = None,
    ) -> User | list[User]:
        if key is not None:
            keys = normalize(key)
        if username is not None:
            usernames = normalize(username)
        return send_required(
            self.client,
            _RETRIEVE_ENDPOINT,
            _RetrieveRequest(keys=keys, usernames=usernames),
            _RetrieveResponse,
        ).users

    def delete(self, keys: UUID | list[UUID] | None = None) -> None:
        send_required(
            self.client,
            _DELETE_ENDPOINT,
            _DeleteRequest(keys=normalize(keys)),
            Empty,
        )



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from datetime import datetime

from pandas import DataFrame
from rich.progress import (
    BarColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
)

from synnax import Synnax
from synnax.channel import Channel
from synnax.framer import Writer
from synnax.io import RowFileReader
from synnax.telem import Size, TimeStamp


class RowIngestionEngine:
    """An ingestion engine that reads data from a row-based reader and writes it to a
    Synnax cluster.
    """

    client: Synnax
    writer: Writer
    reader: RowFileReader
    channels: list[Channel]
    idx_grouped: dict[Channel, list[Channel]]
    end: TimeStamp

    def __init__(
        self,
        client: Synnax,
        reader: RowFileReader,
        channels: list[Channel],
        start: TimeStamp,
        soft_mem_limit: int = 10 * Size.MB,
    ):
        self.channels = channels
        self.idx_grouped = {ch: list() for ch in channels if ch.is_index}
        for ch in self.idx_grouped:
            self.idx_grouped[ch] = [_ch for _ch in channels if _ch.index == ch.key]
        self.mem_limit = soft_mem_limit
        self.reader = reader
        self.client = client
        self.reader.set_chunk_size(self.get_chunk_size())
        self.writer = self.client.open_writer(
            start, [ch.key for ch in channels], err_on_extra_chans=False
        )
        self.end = start

    def get_chunk_size(self):
        """Sum the density of all channels to determine the chunk size."""
        return self.mem_limit // sum(ch.data_type.density for ch in self.channels)

    def run(self):
        """Run the ingestion engine."""
        self.reader.seek_first()
        try:
            with Progress(
                BarColumn(),
                TaskProgressColumn(),
                TextColumn("{task.completed} out of {task.total} samples"),
                TimeElapsedColumn(),
                TextColumn("{task.fields[tp]} samples/s"),
            ) as progress:
                task = progress.add_task("ingest", total=self.reader.nsamples(), tp=0)
                while True:
                    try:
                        t0 = datetime.now()
                        chunk = self.reader.read()
                        self._write(chunk)
                        tp = chunk.size / (datetime.now() - t0).total_seconds()
                        progress.update(task, advance=chunk.size, tp=int(tp))
                    except StopIteration:
                        break
            self.end, _ = self.writer.commit()
        finally:
            self.reader.close()
            self.writer.close()

    def _write(self, df: DataFrame):
        for channel in self.channels:
            if channel.name in df.columns:
                df.rename(columns={channel.name: channel.key}, inplace=True)
                df[channel.key] = df[channel.key].astype(channel.data_type.np)
        self.writer.write(df)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from enum import Enum
from typing import overload, TypeAlias, Literal
from uuid import uuid4
from warnings import warn

from numpy import can_cast as np_can_cast
from pandas import DataFrame
from pandas import concat as pd_concat

from freighter import (
    EOF,
    Payload,
    Stream,
    StreamClient,
    decode_exception,
)
from synnax import io
from synnax.channel.payload import (
    ChannelKey,
    ChannelKeys,
    ChannelName,
    ChannelNames,
    ChannelPayload,
)
from synnax.exceptions import Field, ValidationError
from synnax.framer.adapter import WriteFrameAdapter
from synnax.framer.frame import Frame, FramePayload, CrudeFrame
from synnax.telem import CrudeSeries, CrudeTimeStamp, DataType, TimeSpan, TimeStamp
from synnax.telem.control import Authority, Subject, CrudeAuthority
from synnax.util.normalize import normalize


class _Command(int, Enum):
    OPEN = 0
    WRITE = 1
    COMMIT = 2
    ERROR = 3
    SET_AUTHORITY = 4
    SET_MODE = 5


class WriterMode(int, Enum):
    PERSIST_STREAM = 1
    PERSIST = 2
    STREAM = 3


CrudeWriterMode: TypeAlias = (
    WriterMode
    | Literal["persist_stream"]
    | Literal["persist"]
    | Literal["stream"]
    | int
)


def parse_writer_mode(mode: CrudeWriterMode) -> WriterMode:
    if mode == "persist_stream":
        return WriterMode.PERSIST_STREAM
    if mode == "persist":
        return WriterMode.PERSIST
    if mode == "stream":
        return WriterMode.STREAM
    if isinstance(mode, WriterMode):
        return mode
    if isinstance(mode, int):
        try:
            return WriterMode(mode)
        except:
            raise ValueError(f"invalid writer mode {mode}")


class _Config(Payload):
    authorities: list[int] = Authority.ABSOLUTE
    control_subject: Subject = Subject(name="", key=str(uuid4()))
    start: TimeStamp | None = None
    keys: ChannelKeys
    mode: WriterMode = WriterMode.PERSIST_STREAM
    err_on_unauthorized: bool = False
    enable_auto_commit: bool = False
    auto_index_persist_interval: TimeSpan = 1 * TimeSpan.SECOND


class _Request(Payload):
    config: _Config | None = None
    command: _Command
    frame: FramePayload | None = None


class _Response(Payload):
    command: _Command
    ack: bool
    error: str | None
    end: TimeStamp | None


ALWAYS_INDEX_PERSIST_ON_AUTO_COMMIT: TimeSpan = TimeSpan(-1)


class Writer:
    """A writer is used to write telemetry to a set of channels in time order. It
    should not be constructed directly, and should instead be created using the Synnax
    client's open_writer method.

    The writer is a streaming protocol that is heavily optimized for performance. This
    comes at the cost of increased complexity, and should only be used directly when
    writing large volumes of data (such as recording telemetry from a sensor or
    ingesting data from a file). Simpler methods (such as the frame writer's write
    method) should be used in most cases.

    The protocol is as follows:

    1. The writer is opened with a starting timestamp and a list of channel keys (or
    names). The writer will fail to open if the starting timestamp overlaps with any
    existing telemetry for any of the channels specified. If the writer is opened
    successfully, the caller is then free to write frames to the writer.

    2. To writer a frame, the caller can use the write method and follow the validation
    rules described in its method's documentation. This process is asynchronous, meaning
    that write will return before the frame has been written to the cluster. This also
    means that the writer can accumulate an error after write is called. If the writer
    accumulates an error, all subsequent write and commit calls will return False. The
    caller can check for errors by calling the error method, which returns the
    accumulated error and resets the writer for future use. The caller can also check
    for errors by closing the writer, which will raise any accumulated error.

    3. To commit the written frames to the database, the caller can call the commit
    method. Unlike write, commit is synchronous, meaning that it will not return until
    all frames have been committed to the database. If the writer has accumulated an
    error, commit will return False. After the caller acknowledges the error, they can
    attempt to commit again. Commit can be called several times throughout a writer's
    lifetime, and will commit all frames written since the last commit.

    4. A writer MUST be closed after use in order to prevent resource leaks. Close
    should typically be called in a 'finally' block. If the writer has accumulated an
    error, close will raise the accumulated error.
    """

    __ENDPOINT = "/frame/write"
    __stream: Stream[_Request, _Response]
    __adapter: WriteFrameAdapter
    __suppress_warnings: bool = False
    __strict: bool = False

    start: CrudeTimeStamp

    def __init__(
        self,
        start: CrudeTimeStamp,
        client: StreamClient,
        adapter: WriteFrameAdapter,
        name: str = "",
        authorities: list[Authority] | Authority = Authority.ABSOLUTE,
        suppress_warnings: bool = False,
        strict: bool = False,
        mode: CrudeWriterMode = WriterMode.PERSIST_STREAM,
        err_on_unauthorized: bool = False,
        enable_auto_commit: bool = False,
        auto_index_persist_interval: TimeSpan = 1 * TimeSpan.SECOND,
    ) -> None:
        self.start = start
        self.__adapter = adapter
        self.__suppress_warnings = suppress_warnings
        self.__strict = strict
        self.__stream = client.stream(self.__ENDPOINT, _Request, _Response)
        config = _Config(
            control_subject=Subject(name=name, key=str(uuid4())),
            keys=self.__adapter.keys,
            start=TimeStamp(self.start),
            authorities=normalize(authorities),
            mode=parse_writer_mode(mode),
            err_on_unauthorized=err_on_unauthorized,
            enable_auto_commit=enable_auto_commit,
            auto_index_persist_interval=auto_index_persist_interval,
        )
        self.__stream.send(_Request(command=_Command.OPEN, config=config))
        _, exc = self.__stream.receive()
        if exc is not None:
            raise exc

    @overload
    def write(self, channels_or_data: ChannelName, series: CrudeSeries):
        ...

    @overload
    def write(
        self, channels_or_data: ChannelKeys | ChannelNames, series: list[CrudeSeries]
    ):
        ...

    @overload
    def write(
        self,
        channels_or_data: CrudeFrame,
    ):
        ...

    def write(
        self,
        channels_or_data: (
            ChannelName | ChannelKey | ChannelKeys | ChannelNames | CrudeFrame
        ),
        series: CrudeSeries | list[CrudeSeries] | None = None,
    ) -> bool:
        """Writes the given data to the database. The formats are listed below. Before
        we get into them, here are some important terms to know.

            1. Channel ID -> the key or name of the channel(s) you're writing to.
            2. Series or CrudeSeries -> the data for that channel, which can be
            represented as a synnax Series type, a numpy array, or a simple Python
            list. You can also provide a single numeric (or, in the case of variable
            length types, a string or JSON) value and Synnax will convert it into a
            Series for you.

        Here are the formats you can use to write data to a Synnax cluster:

            1. Channel ID and a single series: Writes the series for the given channel.
            2. A list of channel ids and their corresponding series: Assumes a
            one to one mapping of ids to series i.e. the channel id at index i
            corresponds to the series at index i.
            3. A Synnax Frame (see the Frame documentation for more).
            4. A dictionary of channel ids to series i.e. write the series for the
            given channel id.
            5. A pandas DataFrame where the columns are the channel ids and the rows
            are the series to write.
            6. A dictionary of channel ids to a single
            numeric value. Synnax will convert this into a series for you.

        There are a few important rules to keep in mind when writing data to a Synnax
        cluster:

            1. Have exactly one array for each key in the list of keys provided to the
            writer's open method.
            2. Have equal length arrays for each key.
            3. When writing to an index (i.e. TimeStamp) channel, the values must be
            monotonically increasing.

        :returns: False if the writer has accumulated an error. If this is the case,
        the caller should acknowledge the error by calling the error method or closing
        the writer.
        """
        if self.__stream.received():
            return False

        frame = self.__adapter.adapt(channels_or_data, series)
        self.__check_keys(frame)
        self.__prep_data_types(frame)

        exc = self.__stream.send(
            _Request(command=_Command.WRITE, frame=frame.to_payload())
        )
        if exc is not None:
            raise exc
        return True

    @overload
    def set_authority(self, value: CrudeAuthority) -> bool:
        ...

    @overload
    def set_authority(
        self,
        value: ChannelKey | ChannelName,
        authority: CrudeAuthority,
    ) -> bool:
        ...

    @overload
    def set_authority(
        self,
        value: dict[ChannelKey | ChannelName | ChannelPayload, CrudeAuthority],
    ) -> bool:
        ...

    def set_authority(
        self,
        value: (
            dict[ChannelKey | ChannelName | ChannelPayload, CrudeAuthority]
            | ChannelKey
            | ChannelName
            | CrudeAuthority
        ),
        authority: CrudeAuthority | None = None,
    ) -> bool:
        if isinstance(value, int) and authority is None:
            cfg = _Config(keys=[], authorities=[value])
        else:
            if isinstance(value, (ChannelKey, ChannelName)):
                if authority is None:
                    raise ValueError(
                        "authority must be provided when setting a single channel"
                    )
                value = {value: authority}
            value = self.__adapter.adapt_dict_keys(value)
            cfg = _Config(
                keys=list(value.keys()),
                authorities=list(value.values()),
            )
        exc = self.__stream.send(_Request(command=_Command.SET_AUTHORITY, config=cfg))
        if exc is not None:
            raise exc
        while True:
            res, exc = self.__stream.receive()
            if exc is not None:
                raise exc
            if res.command == _Command.SET_AUTHORITY:
                return res.ack

    def commit(self) -> tuple[TimeStamp, bool]:
        """Commits the written frames to the database. Commit is synchronous, meaning
        that it will not return until all frames have been committed to the database.

        :returns: False if the commit failed due to an error. In this case, the caller
        should acknowledge the error by calling the error method or closing the writer.
        After the error is acknowledged, the caller can attempt to commit again.
        """
        if self.__stream.received():
            return TimeStamp.ZERO, False
        exc = self.__stream.send(_Request(command=_Command.COMMIT))
        if exc is not None:
            raise exc

        while True:
            res, exc = self.__stream.receive()
            if exc is not None:
                raise exc
            if res.command == _Command.COMMIT:
                return res.end, res.ack

    def error(self) -> Exception | None:
        """
        :returns: The exception that the writer has accumulated, if any. If the writer
        has not accumulated an error, this method will return None. This method will
        clear the writer's error state, allowing the writer to be used again.
        """
        self.__stream.send(_Request(command=_Command.ERROR))

        while True:
            res, exc = self.__stream.receive()
            if exc is not None:
                raise exc
            assert res is not None
            if res.command == _Command.ERROR:
                return decode_exception(res.error)

    def close(self):
        """Closes the writer, raising any accumulated error encountered during
        operation. A writer MUST be closed after use, and this method should probably
        be placed in a 'finally' block.
        """
        self.__stream.close_send()
        while True:
            res, exc = self.__stream.receive()
            if exc is None:
                assert res is not None
                exc = decode_exception(res.error)

            if exc is not None:
                if isinstance(exc, EOF):
                    return
                raise exc

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __check_keys(self, frame: Frame):
        missing = set(self.__adapter.keys) - set(frame.channels)
        extra = set(frame.channels) - set(self.__adapter.keys)
        if missing and extra:
            raise ValidationError(
                Field(
                    "keys",
                    f"frame is missing keys {missing} and has extra keys {extra}",
                )
            )
        elif extra:
            raise ValidationError(Field("keys", f"frame has extra keys {extra}"))

    def __prep_data_types(self, frame: Frame):
        for i, (col, series) in enumerate(frame.items()):
            ch = self.__adapter.retriever.retrieve(col)[0]  # type: ignore
            if series.data_type != ch.data_type:
                if (
                    not np_can_cast(series.data_type.np, ch.data_type.np)
                    or self.__strict
                ):
                    raise ValidationError(
                        Field(
                            str(col),
                            f"""Column {col} has type {series.data_type} but channel
                            {ch.key} expects type {ch.data_type}""",
                        )
                    )
                elif not self.__suppress_warnings and not (
                    ch.data_type == DataType.TIMESTAMP
                    and series.data_type == DataType.INT64
                ):
                    warn(
                        f"""Series for channel {ch.name} has type {series.data_type} but
                        channel expects type {ch.data_type}. We can safely convert
                        between the two, but this can cause performance degradations
                        and is not recommended. To suppress this warning,
                        set suppress_warnings=True when constructing the writer. To
                        raise an error instead, set strict=True when constructing
                        the writer."""
                    )
                frame.series[i] = series.astype(ch.data_type)


class BufferedWriter(io.DataFrameWriter):
    """BufferedWriter extends the Writer class by buffering
    writes to the underlying stream. This can improve performance by reducing the
    number of round trips to the server.
    """

    size_threshold: int
    time_threshold: TimeSpan
    last_flush: TimeStamp
    _wrapped: Writer
    _buf: DataFrame

    def __init__(
        self,
        wrapped: Writer,
        size_threshold: int = int(1e6),
        time_threshold: TimeSpan = TimeSpan.MAX,
    ) -> None:
        self._wrapped = wrapped
        self._buf = DataFrame()
        self.last_flush = TimeStamp.now()
        self.size_threshold = size_threshold
        self.time_threshold = time_threshold

    def _(self) -> io.DataFrameWriter:
        return self

    def write(self, frame: DataFrame):
        self._buf = pd_concat([self._buf, frame], ignore_index=True)
        if self._exceeds_any:
            self.__flush()

    def close(self):
        self.__flush()
        self._wrapped.close()

    @property
    def _exceeds_any(self) -> bool:
        return (
            len(self._buf) * len(self._buf.columns) >= self.size_threshold
            or TimeSpan.since(self.last_flush) >= self.time_threshold
        )

    def __flush(self):
        self._wrapped.write(self._buf)
        self._wrapped.commit()
        self.last_flush = TimeStamp.now()
        self._buf = DataFrame()



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations
from freighter import Payload


class ID(Payload):
    key: str | None = ""
    type: str | None = ""

    def __init__(self, key: CrudeID | None = None, type: str | None = None):
        if isinstance(key, ID):
            super().__init__(key=key.key, type=key.type)
        elif isinstance(key, tuple):
            key, type = key
            super().__init__(key=key, type=type)
        elif type is None:
            type, key = key.split(":")
            super().__init__(key=key, type=type)
        else:
            super().__init__(key=key, type=type)


ROOT = ID(key="root", type="builtin")

CrudeID = str | ID


class Resource(Payload):
    id: ID
    name: str
    data: dict


class Relationship(Payload):
    from_: ID
    type: str
    to: ID
